from arm64_tester.parameters.array_parameter import ArrayParameter
from arm64_tester.parameters.numeric_parameter import NumericParameter
from arm64_tester.parameters.string_parameter import StringParameter


__all__ = 'CompileError'


class CompileError(Exception):

    def __init__(self, message):
        first_index = str(message).find('"') + 1
        last_index = str(message).rfind('"')

        self.message = str(message)[
            first_index:last_index].strip()
        super().__init__(self.message)



from ast import literal_eval

from arm64_tester.parameters import ArrayParameter as Array
from arm64_tester.parameters import NumericParameter as Numeric
from arm64_tester.parameters import StringParameter as String
from arm64_tester.subroutines.subroutine import Subroutine


class MixedSubroutine(Subroutine):
    """Subroutine that returns both a number and one or more arrays"""

    def __init__(self, name, parameters, number_return_type, array_outputs):
        super().__init__(name, parameters)
        self.c_function_return = number_return_type
        if number_return_type == 'int':
            self.printf_format = 'd'
        elif number_return_type == 'long':
            self.printf_format = 'ld'
        elif number_return_type == 'float' or number_return_type == 'double':
            self.printf_format = 'f'
        elif number_return_type == 'char':
            self.printf_format = 'c'
        self.array_outputs = array_outputs

    def get_nr_outputs(self):
        return len(self.array_outputs)

    def build_test_call(self):
        return '{} {} {} printf("\\n");'.format(\
            # Declare output variables beforehand, so we have access to them after subroutine call
            ''.join([parameter.get_test_declaration_representation() for parameter in self.parameters]),\
            # Actually make subroutine call
            'printf("%{}\\n",{}({}));'.format(self.printf_format, self.name, ','.join([parameter.get_test_call_representation() for parameter in self.parameters])),\
            # Access previously declared variables to print their final values
            'printf("\\n");'.join(filter(lambda x: x != '', [parameter.get_test_call_output_representation() for parameter in self.parameters])))

    def process_parameters(self, parameters):
        for idx, parameter in enumerate(parameters):
            if parameter == 'string':
                self.parameters.append(String(idx, True if idx >= (
                    len(parameters) - len(self.array_outputs)) else False))
            elif 'array' in parameter:
                self.parameters.append(Array(idx, parameter.replace('array', '').strip(
                ), True if idx >= (len(parameters) - len(self.array_outputs)) else False))
            else:  # numeric
                self.parameters.append(Numeric(idx, parameter))

    def compare_outputs(self, expected, real, precision):
        if(len(expected) != len(real)):
            return False
        if (self.printf_format == 'd' or self.printf_format == 'ld') and expected[0] != int(real[0]):
            return False
        elif self.printf_format == 'c':
            if expected[0] is None and real[0].rstrip('\x00') != "" and real[0] != '0':
                return False
            elif expected[0] is not None and expected[0] != real[0]:
                return False
        elif self.printf_format == 'f' and abs(expected[0]-float(real[0])) > precision:
            return False

        for out_type, exp, re in zip(self.array_outputs, expected[1:], real[1:]):
            if out_type == 'string':
                if exp != re:
                    return False
            else:  # Array
                arr_type = out_type.replace('array', '').strip()
                re_arr = literal_eval(re)
                if(len(exp) != len(re_arr)):
                    return False
                for exp_el, re_el in zip(exp, re_arr):
                    if (arr_type == 'int' or arr_type == 'long') and exp_el != re_el:
                        return False
                    elif abs(exp_el-re_el) > precision:
                        return False

        return True



from datetime import datetime
from re import IGNORECASE, match

from yaml import YAMLError, safe_load

from arm64_tester.exception import ToolFileError
from arm64_tester.subroutines import ArraySubroutine as Array
from arm64_tester.subroutines import MixedSubroutine as Mixed
from arm64_tester.subroutines import NumericSubroutine as Numeric
from arm64_tester.subroutines import VoidSubroutine as Void
from arm64_tester.tester import Tester


class Test:

    def __init__(self, submission_file, subroutine_file, tests_file,
                 timeout=1.0, float_threshold=1e-6, save_to_file=False):
        self.submission_file = submission_file
        self.subroutine_file = subroutine_file
        self.tests_file = tests_file
        self.timeout = timeout
        self.float_threshold = float_threshold
        self.save_to_file = save_to_file

        self.__read_files()
        self.__build_test_cases()

    def __read_files(self):
        try:
            self.subroutines = safe_load(open(self.subroutine_file))
            self.test_suite = safe_load(open(self.tests_file))
        except IOError as err:
            raise ToolFileError(
                'Could not open: {}, please specify a valid file'.format(str(err)))
        except YAMLError as err:
            raise ToolFileError(
                'Error parsing YAML files: ({}), please correct syntax'.format(str(err)))

    def __build_subroutine(self, name, definition):
        map_to_truth_value = ['array' in ret or ret ==
                              'string' for ret in definition['return']]
        if not len(map_to_truth_value):  # No returns - void subroutine
            return Void(name, definition['params'])
        elif all(map_to_truth_value):  # Only arrays and/or strings as return values
            return Array(name, definition['params'], definition['return'])
        elif not any(map_to_truth_value):  # Numeric output
            return Numeric(name, definition['params'], definition['return'][0])
        else:  # Mixed return
            return Mixed(name, definition['params'], definition['return'][0], definition['return'][1:])

    def __build_test_cases(self):
        self.subroutine_objects = {name: self.__build_subroutine(
            name, definition) for name, definition in self.subroutines.items()}

    def run(self):
        folder_prefix = str(datetime.now().microsecond)

        grading_folder = f'tmp_{folder_prefix}_grading'
        feedback_folder = f'tmp_{folder_prefix}_feedback'

        tester = Tester(self.subroutine_objects, self.test_suite, grading_folder=grading_folder,
                        feedback_folder=feedback_folder, float_threshold=self.float_threshold,
                        timeout=self.timeout, save_to_file=self.save_to_file)

        submission = tester.grade_submission(self.submission_file)

        filename = match(
            r'(?:.+\/)*(.*)?.*\.zip', self.submission_file, flags=IGNORECASE).group(1)

        """ subroutines_result = [{
            'name': subroutine['name'],
            'score': subroutine['score']
        } for subroutine in submission] """

        return {
            'submission_name': filename,
            'subroutines': submission,
        }



class Parameter:
    """Abstract parent class that defines a generic subroutine parameter"""

    def __init__(self, idx):
        # By default, not a function output (to be overriden e.g., by arrays/strings)
        self.is_output = False
        self.idx = idx  # Index to uniquely identify parameter amongst all parameters

    def get_prototype_representation(self):
        """Method that returns argument's representation as a function argument in its prototype"""
        pass

    def get_test_declaration_representation(self):
        """Method that returns parameter's declaration to use before a function call, when it is of the adequate type"""
        pass

    def get_test_call_representation(self):
        """Method that returns parameter's representation as an actual argument to a subroutine call (may be direct value or variable, where value is not used)"""
        pass

    def get_literal_representantion(self, value):
        """Method that returns parameter's representation for when to format an input into a template C file"""
        pass

    def get_test_call_output_representation(self):
        """Method that calculates the adequate printf statements post-subroutine call for arrays and strings"""
        pass



from arm64_tester.parameters.parameter import Parameter


class StringParameter(Parameter):
    """String subroutine parameter"""

    def __init__(self, idx, is_output):
        super().__init__(idx)
        self.is_output = is_output

    def get_prototype_representation(self):
        return 'char* arg{}'.format(self.idx)

    def get_test_declaration_representation(self):
        return 'char arg{}[] = "{{}}";'.format(self.idx) if self.is_output else ''

    def get_test_call_representation(self):
        return 'arg{}'.format(self.idx) if self.is_output else '"{}"'

    def get_literal_representantion(self, value):
        return value

    def get_test_call_output_representation(self):
        return 'printf("%s", arg{});'.format(self.idx) if self.is_output else ''



from arm64_tester.subroutines.array_subroutine import ArraySubroutine
from arm64_tester.subroutines.mixed_subroutine import MixedSubroutine
from arm64_tester.subroutines.numeric_subroutine import NumericSubroutine
from arm64_tester.subroutines.void_subroutine import VoidSubroutine



import json
from argparse import ArgumentParser
from re import IGNORECASE, match
from sys import exit

from yaml import YAMLError, safe_load

from arm64_tester.subroutines import ArraySubroutine as Array
from arm64_tester.subroutines import MixedSubroutine as Mixed
from arm64_tester.subroutines import NumericSubroutine as Numeric
from arm64_tester.subroutines import VoidSubroutine as Void
from arm64_tester.tester import Tester as Evaluator


def parse_args():
    """Uses argparse module to create a pretty CLI interface that has the -h by default and that helps the user understand the arguments and their usage
    Returns a dict of "argname":"value"
    """
    parser = ArgumentParser(
        prog="code-correction",
        description="AOCO - Automatic Observation and Correction of (subroutine) Operations")

    parser.add_argument(
        '-sr', help='YAML file containing definition of subroutines', required=True)
    parser.add_argument(
        '-t', help='YAML file containing test cases for evaluation', required=True)
    parser.add_argument(
        '-sm', help='Whitespace separated list of .zip files corresponding to students\' submissions', required=True, nargs='+')
    parser.add_argument(
        '-gfd', help='Folder to store temporary files necessary for grading a submission. Defaults to \'grading\'', default='grading')
    parser.add_argument(
        '-ffd', help='Folder to store submission feedback for each student. Defaults to \'feedback\'', default='feedback')
    parser.add_argument(
        '-grf', help='CSV file to store student grades. Defaults to \'grades.csv\'', default='grades.csv')
    parser.add_argument(
        '-tout', help='Timeout value for when running compiled submissions. Defaults to 2 (seconds)', type=float, default=2)
    parser.add_argument(
        '-fpre', help='Floating point threshold, for when comparing floating point number outputs (accepts values within [VALUE-threshold,VALUE+threshold]). Defaults to 1e-6', type=float, default=1e-6)

    return vars(parser.parse_args())


def build_subroutine(name, definition):
    """Creates a C file template that will run all the test inputs for a given subroutine later
    """
    map_to_truth_value = ['array' in ret or ret ==
                          'string' for ret in definition['return']]
    if not len(map_to_truth_value):  # No returns - void subroutine
        return Void(name, definition['params'])
    elif all(map_to_truth_value):  # Only arrays and/or strings as return values
        return Array(name, definition['params'], definition['return'])
    elif not any(map_to_truth_value):  # Numeric output
        return Numeric(name, definition['params'], definition['return'][0])
    else:  # Mixed return
        return Mixed(name, definition['params'], definition['return'][0], definition['return'][1:])


if __name__ == "__main__":
    args = parse_args()
    try:
        subroutines = safe_load(open(args['sr']))
        test_suite = safe_load(open(args['t']))
    except IOError as err:
        print('Could not open: {}, please specify a valid file'.format(str(err)))
        exit(-1)
    except YAMLError as err:
        print('Error parsing YAML files: ({}), please correct syntax'.format(str(err)))

    subroutine_objs = {name: build_subroutine(
        name, definition) for name, definition in subroutines.items()}

    program_args = {key: args[key]
                    for key in ['gfd', 'ffd', 'grf', 'tout', 'fpre']}
    evaluator = Evaluator(subroutine_objs, test_suite,
                          cmd_line_args=program_args)

    result = []

    for student_submission in args['sm']:
        submission = evaluator.grade_submission(student_submission)

        filename = match(
            r'(?:.+\/)*(.*)?.*\.zip', student_submission, flags=IGNORECASE).group(1)

        subroutines_result = [{
            'name': subroutine['name'],
            'score': subroutine['score']
        } for subroutine in submission]

        result.append({
            'submission_name': filename,
            'subroutines': subroutines_result
        })

    file = open('{}/result.json'.format(program_args['ffd']), 'w')
    json.dump(result, file, indent=4)
